Level = coup.class({})

local TilePosUpdater = require "tileposupdater"
local BlockGravityUpdater = require "blockgravityupdate"
local CoinChecker = require "coin_checker"
local HintChecker = require "hint_checker"
local Particles = require "particles"

function loadTiles(name)
  local filename = string.gsub(name, "../", "data/", 1)
  return G.loadImage(filename)
end


function Level:createObjects(objs)
  local objects = objs.objects.objects
  for i, v in ipairs(objects) do
    local x = v.x / self.tilemap.tilewidth
    local y = v.y / self.tilemap.tileheight
    if v.type == "dwarf" then
      local e = self.world:newEntity("dwarf")
      e:makeComponent("TilePosition", x, y)
      e:makeComponent("Sprite",      G.animRepo:newSprite("Idle_R")).layer = 2
      e:makeComponent("BoundingBox", -32, -32, 32, 32)
      e:makeComponent("PlayerState")
    elseif v.type == "gold" then
      local e = self.world:newEntity()
      e:makeComponent("TilePosition",  x, y)
      local spr = e:makeComponent("Sprite",      G.animRepo:newSprite("Gold_Idle"))
      spr.sprite.time = math.random(0, 50)
      e:makeComponent("BoundingBox", -32, -32, 32, 32)
      e:makeComponent("GoldCoin")
    elseif v.type == "exit" then
      local e = self.world:newEntity("exit")
      e:makeComponent("TilePosition",  x, y)
      e:makeComponent("Sprite",      G.animRepo:newSprite("Exit_Closed"))
      e:makeComponent("BoundingBox", -32, -32, 32, 32)
    elseif v.type == "block" then
      local e = self.world:newEntity()
      e:makeComponent("TilePosition",  x, y)
      e:makeComponent("Sprite",      G.animRepo:newSprite("Block_Idle")).layer=1
      e:makeComponent("BoundingBox", -32, -32, 32, 32)
      e:makeComponent("PushableBlock")
    elseif v.type == "hint" then
      local h = self.world:newEntity()
      h:makeComponent("TilePosition", x, y)
      h:makeComponent("Hint", v.properties["title"], v.properties["text"])
    elseif v.type == "elevator" then
      local h = self.world:newEntity()
      h:makeComponent("TilePosition", x, y)
      local ymax = y
      while ymax < 9 and not self:isSolid(x, ymax+1) do
        ymax = ymax + 1
      end

      local ymin = y
      while ymin > 0 and not self:isSolid(x, ymin-1) do
        ymin = ymin -1 
      end

      local tracks = {}
      for i = ymin, ymax do
        local t = self.world:newEntity()
        t:makeComponent("Sprite", G.animRepo:newSprite("Tracks_Idle"))
        t:makeComponent("TilePosition", x, i)
        t:makeComponent("BoundingBox", -32, -32, 32, 32)
        tracks[i] = t
      end

      local engine = self.world:newEntity()
      engine:makeComponent("Sprite", G.animRepo:newSprite("Engine_Idle"))
      engine:makeComponent("TilePosition", x, ymin-1)
      engine:makeComponent("BoundingBox", -32, -32, 32, 32) 


      h:makeComponent("Sprite",      G.animRepo:newSprite("Elevator_Idle")).layer=1
      h:makeComponent("BoundingBox", -32, -32, 32, 32)
      h:makeComponent("Elevator", tracks, engine)
    end
  end
end


function Level:new(name)
  self.hint = {}
  self.timer = 0
  self.vignette = G.loadImage("data/vignette.png")
  self.shake = 0
  self.filename = name
  self.turns = 0
  self.time = 0
  self.ui_tween = coup.tween.group()
  self.tween = coup.tween.group()
  self.coins = 0
  self.fast_anim = false
  self.world = coup.world.World()
  self.particles = Particles()
  coup.components.register(self.world)
  self.world:registerComponentType("TilePosition", require("tileposition")())
  self.world:registerComponentType("TileMovement", require("tilemovement")())
  self.world:registerComponentType("PlayerState", require("playerstate")())
  self.world:registerComponentType("PushableBlock", require("pushable_block")())
  self.world:registerComponentType("GoldCoin", require("gold_coin")())
  self.world:registerComponentType("OpenExit", require("open_exit")())
  self.world:registerComponentType("Hint", require("hint")())
  self.world:registerComponentType("Elevator", require("elevator")())

  local objs, imgs, props
  self.tilemap, objs, imgs, props = coup.tiledloader.loadTiledMapFile("data/levels/" .. name .. ".lua", loadTiles)

  self.nextLevel = props["next"]
  self.name = props["name"]
  self.mainLayer = self.tilemap:get_layer("main")
  self.mainLayerRenderer = coup.tilelayerrenderer.TilelayerRenderer(self.mainLayer, 10)

  self.bgLayer = self.tilemap:get_layer("background")
  self.bgLayerRenderer = coup.tilelayerrenderer.TilelayerRenderer(self.bgLayer, 10)

  self:createObjects(objs)

  self.spriteSpatialHash     = coup.spatialhash.SpatialHashtable(50)
  self.spriteAutoHasher      = coup.systems.AutoHasher(self.spriteSpatialHash, self.world, "Sprite")
  self.aabbUpdater           = coup.systems.AABBUpdater(self.world)
  self.spriteRenderer        = coup.systems.SpriteRenderer(self.spriteSpatialHash)
  self.spriteUpdater         = coup.systems.SpriteUpdater(self.world)
  self.tilePosUpdater        = TilePosUpdater(self.world, self.tilemap.tilewidth, self.tilemap.tileheight)
  self.blockGravityUpdater   = BlockGravityUpdater(self)
  self.coinChecker           = CoinChecker(self)
  self.hintChecker           = HintChecker(self)

  self.camera = self.world:newEntity("camera")
  local w, h = love.graphics.getDimensions()
  self.camera:makeComponent("Transform", w/2, h/2)
  self.camera:makeComponent("BoundingBox", -500, -500, 500, 500)

  self.world:step()
  self:update(0)

  self.allowInput = true
  self.lastDust = 0
  self.lastSkit = 0
end


function Level:setPlayerAnim()
  local dwarf = self.world:getEntityByName("dwarf")

  local p = dwarf:getComponent("TilePosition")
  local m = dwarf:getComponent("TileMovement")
  local s = dwarf:getComponent("PlayerState")
  local spr = dwarf:getComponent("Sprite")

  if s.dir == "l" or s.dir == "r" then
    if m then
      if m.pos.y == p.pos.y then
        if m.pos.x < p.pos.x then
          s.dir = "l"
          spr.sprite:play("Walk_L", true)
        elseif m.pos.x > p.pos.x then
          s.dir = "r"
          spr.sprite:play("Walk_R", true)
        end
        local t = dwarf:getComponent("Transform")
        if self.lastDust < self.time - 0.1 then
          local c = math.random(64, 192)
          self.particles:spawn(t.pos.x+48, t.pos.y+64, math.random(-10, 10), math.random(-40, -20), 0, math.random(1,3), 0, 7.5, 2, 2, c,c,c)
          self.lastDust = self.time
        end
        if self.lastSkit < self.time - 0.3 then
          G.sounds.skit:stop()
          G.sounds.skit:play()
          self.lastSkit = self.time
        end
      end
    else
      if s.dir == "l" then
        spr.sprite:play("Idle_L", true)
      else
        spr.sprite:play("Idle_R", true)
      end
    end
  end
end


function Level:update(dt)


  self.ui_tween:update(dt)
  self.time = self.time + dt
  if self:inputAllowed() and self.eventQueue then
    if self.queueTime > self.time - 0.2 then
      self:inputEvent(self.eventQueue)
    end
    self.eventQueue = nil
  end

  if G.fast_anim then
    dt = dt * 2
  end

  self.tween:update(dt)
  self:setPlayerAnim()
  self.blockGravityUpdater:update()
  self.tilePosUpdater:update(dt)
  self.hintChecker:update()
  self.aabbUpdater:update()
  self.spriteAutoHasher:update()
  self.spriteUpdater:updateSprites(dt)
  self.world:step()
  self.particles:update(dt)
end


function Level:showHint(title, text, func)
  self.hint = {title = title, text = text, x = 640, y = 80, alpha = 0, func = func}
  self.ui_tween:tween(self.hint, 0.4, {x = 80, alpha = 1.0})
end


function Level:screenshake(shake, duration)
  self.shake = shake or 20
  self.ui_tween:tween(self, duration or 0.3, {shake = 0})
end


function Level:render()
  local ms = math.floor(self.shake)
  local sx = math.random(-ms, ms)
  local sy = math.random(-ms, ms)

  love.graphics.push()
  love.graphics.translate(sx, sy)

  local cab = self.camera:getComponent("AxisAlignedBoundingBox")
  self.bgLayerRenderer:draw(cab.topLeft, cab.bottomRight)
  self.mainLayerRenderer:draw(cab.topLeft, cab.bottomRight)
  self.spriteRenderer:renderVisible(self.camera)
  self.particles:render()
  love.graphics.pop()

  love.graphics.setBlendMode("multiply")
  love.graphics.draw(self.vignette)
  love.graphics.setBlendMode("alpha")

  if self.hint.title then
    love.graphics.setFont(G.titleFont)
    love.graphics.setColor(0,0,0,192*self.hint.alpha)
    love.graphics.rectangle("fill", self.hint.x, self.hint.y, 460, 460)
    love.graphics.setColor(119,43,77,255*self.hint.alpha)
    love.graphics.printf(self.hint.title, self.hint.x+20, self.hint.y, 420, "center")
    love.graphics.setColor(240,240,240,255*self.hint.alpha)
    love.graphics.printf(self.hint.text, self.hint.x+20, self.hint.y+50, 420, "left")
    love.graphics.setColor(245, 255, 90, 255*self.hint.alpha)
    love.graphics.printf("< Okay >", self.hint.x+20, self.hint.y+430, 420, "center")
    love.graphics.setColor(255,255,255,255)
  end

  love.graphics.setColor(245, 255, 90, 128)

  love.graphics.setFont(G.titleFont)
  love.graphics.printf(self.name, 40, 600, 580, "right")

  love.graphics.setColor(255,255,255,255)

end


function Level:isSolid(x, y)
  return self.mainLayer:get(x, y) ~= 0
end


function Level:inputAllowed()
  local dw = self.world:getEntityByName("dwarf")

  local mc = 0
  self.world:foreachEntityWithComponents(function(e) mc = mc + 1 end, "TileMovement")
  local dir = dw:getComponent("PlayerState").dir

  return self.allowInput and (dir == "l" or dir == "r") and mc == 0
end


function Level:isBlock(x, y)
  local found = nil
  self.world:foreachEntityWithComponents(function(e)
    local tp = e:getComponent("TilePosition")
    if tp.pos.x == x and tp.pos.y == y then
      found = e
    end
  end, "PushableBlock", "TilePosition")

  return found
end

function Level:jump_dust()
  self:dust(0.66)
end 


function Level:climb_dust()
  self:dust(1.403, 2)
end


function Level:deep_dust()
  self:dust(1.6, 2)
end

function Level:dust(t, shake)
  G.level.timer = 1
  G.level.tween:tween(G.level, t, {timer = 0}):on_complete(function()
    local dw = G.level.world:getEntityByName("dwarf")
    local tm = dw:getComponent("TilePosition")
    local tx = tm.pos.x * 64
    local ty = tm.pos.y * 64
    self:screenshake(shake or 5, 0.1)
    for i = 1, 20 do
      local c = math.random(64, 192)
      G.level.particles:spawn(
        tx+32+math.random(0,16), ty+64, --p
        math.random(200, 300), math.random(-10, 10), -- v
        0, math.random(-20,-10), -- a
        0,  -- s
        7.5, -- ds
        math.random(1, 10) / 7, -- life
        {10, 0}, -- damping
        c,c,c)

      G.level.particles:spawn(
        tx+32-math.random(0,16), ty+64, --p
        math.random(-300, -200), math.random(-10, 10), -- v
        0, math.random(-20,-10), -- a
        0,  -- s
        7.5, -- ds
        math.random(1, 10) / 7, -- life
        {10, 0}, -- damping
        c,c,c)
    end
  end)

end


function Level:isElevator(x, y)
  local found = nil
  self.world:foreachEntityWithComponents(function(e)
    local tp = e:getComponent("TilePosition")
    if tp.pos.x == x and tp.pos.y == y then
      found = e
    end
  end, "Elevator", "TilePosition")

  return found
end


function Level:isFloor(x, y)
  return self:isBlock(x, y) or self:isSolid(x, y) or self:isElevator(x, y)
end


function Level:inputEvent(event)
  if event == "restart" then
    G.level = Level(self.filename)
  elseif event == "skip_level" then
    G.level = Level(self.nextLevel)
  end

  if self.hint.title then
    if event == "confirm" then
      self.ui_tween:tween(self.hint, 0.4, {alpha = 0, x = -420}):on_complete(function()
        if self.hint.func then
          self.hint.func()
        end
        self.hint.title = nil
      end)
    end
  else 

    if not self:inputAllowed() then
      self.eventQueue = event
      self.queueTime = self.time
      return
    end

    local dwarf = self.world:getEntityByName("dwarf")
    local pos = dwarf:getComponent("TilePosition")
    local st = dwarf:getComponent("PlayerState")
    local spr = dwarf:getComponent("Sprite")

    self.turns = self.turns + 1

    if event == "pickup" then
      self.coinChecker:update()

      local elev = self:isElevator(pos.pos.x, pos.pos.y)
      if elev then
        local tr = elev:getComponent("Elevator")
        if tr.tracks[pos.pos.y+1] and not (self:isBlock(pos.pos.x, pos.pos.y+1) or self:isElevator(pos.pos.x, pos.pos.y+1)) then
          elev:makeComponent("TileMovement", pos.pos.x, pos.pos.y + 1, 1)
          dwarf:makeComponent("TileMovement", pos.pos.x, pos.pos.y + 1, 1)
          --G.sounds.elevator:play()
          play_elevator()
          if st.dir == "l" then
            spr.sprite:play("Idle_L")
          else
            spr.sprite:play("Idle_R")
          end
          tr.engine:getComponent("Sprite").sprite:play("Engine_Up")
          local i = 1
          while true do
            local bl = self:isBlock(pos.pos.x, pos.pos.y-i)
            if bl then
              bl:makeComponent("TileMovement", pos.pos.x, pos.pos.y-i+1, 1)
            else
              break
            end
            i = i + 1
          end
        end
      end
    elseif event == "up" then
      local exit = self.world:getEntityByName("exit")
      local ee = exit:getComponent("OpenExit")
      local elev = self:isElevator(pos.pos.x, pos.pos.y)
      if ee then
        local ep = exit:getComponent("TilePosition")
        if ep.pos.x == pos.pos.x and ep.pos.y == pos.pos.y then
          spr.sprite:play("Exit_Leave")
          G.sounds.exit:play()
          st.dir = "exit"
          spr.sprite:onEnd(function()
            spr.sprite:play("Exit_Left")
            self:showHint("Level Complete", "You cleared the mine from all its gold and made it to the exit.\n\nYou finished in " .. self.turns .. " turns.", function() G.level = Level(self.nextLevel) end)
          end)
        end
      end
      
      if elev then
        local tr = elev:getComponent("Elevator")
        if tr.tracks[pos.pos.y-1] then
          local cy = pos.pos.y-1
          while self:isBlock(pos.pos.x, cy) do
            cy = cy - 1
          end
          
          if not (self:isSolid(pos.pos.x, cy) or self:isElevator(pos.pos.x, cy)) then
              elev:makeComponent("TileMovement", pos.pos.x, pos.pos.y - 1, 1)
              dwarf:makeComponent("TileMovement", pos.pos.x, pos.pos.y - 1, 1)
              play_elevator()
              if st.dir == "l" then
                spr.sprite:play("Idle_L")
              else
                spr.sprite:play("Idle_R")
              end
              tr.engine:getComponent("Sprite").sprite:play("Engine_Down")
              for i = pos.pos.y-1, cy+1, -1 do
                local bl = self:isBlock(pos.pos.x, i)
                bl:makeComponent("TileMovement", pos.pos.x, i-1, 1)
              end
          end
        end
      else

        if not self:isFloor(pos.pos.x, pos.pos.y-1) then
          if st.dir == "r" then
            if self:isElevator(pos.pos.x + 1, pos.pos.y - 1) 
                or (self:isFloor(pos.pos.x + 1, pos.pos.y) 
                  and not self:isFloor(pos.pos.x + 1, pos.pos.y - 1)) then
              st.dir = "cr"
              spr.sprite:play("Climb_R")
              self:climb_dust()
              --G.sounds.climp:play()
              play_climb()
              pos.pos.y = pos.pos.y - 1
              pos.pos.x = pos.pos.x + 1
              spr.sprite:onEnd(function() 
                st.dir = "r"
                spr.sprite:play("Idle_R")
                local tr = dwarf:getComponent("Transform")
                tr.pos.x = tr.pos.x +48 
              end)
            end
          elseif st.dir == "l" then
            if self:isElevator(pos.pos.x - 1, pos.pos.y - 1) 
                or (self:isFloor(pos.pos.x - 1, pos.pos.y) 
                  and not self:isFloor(pos.pos.x - 1, pos.pos.y - 1)) then
              st.dir = "cl"
              spr.sprite:play("Climb_L")
              self:climb_dust()
              play_climb()
              --G.sounds.climp:play()
              pos.pos.y = pos.pos.y - 1
              pos.pos.x = pos.pos.x - 1
              spr.sprite:onEnd(function() 
                st.dir = "l"
                spr.sprite:play("Idle_L")
                local tr = dwarf:getComponent("Transform")
                tr.pos.x = tr.pos.x -16
              end)
            end
          end
        end
      end

    elseif event == "jump" then

      if st.dir == "r" then
        if not self:isFloor(pos.pos.x + 1, pos.pos.y) and (self:isElevator(pos.pos.x+2, pos.pos.y) or (not self:isFloor(pos.pos.x + 2, pos.pos.y))
            and self:isFloor(pos.pos.x + 2, pos.pos.y+1)) then

          st.dir = "jr"
          spr.sprite:play("Jump_R")
          self:jump_dust()
          G.sounds.jump:stop()
          G.sounds.jump:play()
          pos.pos.x = pos.pos.x + 2
          spr.sprite:onEnd(function() 
            st.dir = "r"
            spr.sprite:play("Idle_R")
            local tr = dwarf:getComponent("Transform")
            tr.pos.x = tr.pos.x +112
          end)

        end
      elseif st.dir == "l" then
        if not self:isFloor(pos.pos.x - 1, pos.pos.y) and (self:isElevator(pos.pos.x-2, pos.pos.y) or (not self:isFloor(pos.pos.x - 2, pos.pos.y))
            and self:isFloor(pos.pos.x - 2, pos.pos.y+1)) then

          st.dir = "jl"
          spr.sprite:play("Jump_L")
          self:jump_dust()
          G.sounds.jump:stop()
          G.sounds.jump:play()
          pos.pos.x = pos.pos.x - 2
          spr.sprite:onEnd(function() 
            st.dir = "l"
            spr.sprite:play("Idle_L")
            local tr = dwarf:getComponent("Transform")
            tr.pos.x = tr.pos.x -16
          end)
        end
      end

    elseif event == "right" then
      if st.dir == "r" then
        local b = self:isBlock(pos.pos.x + 1, pos.pos.y)
        if b then
          if not self:isFloor(pos.pos.x + 2, pos.pos.y) and not self:isBlock(pos.pos.x + 1, pos.pos.y - 1) then
            spr.sprite:play("Push_R")
            pos.pos.x = pos.pos.x + 1
            G.sounds.push:stop()
            G.sounds.push:play()
            spr.sprite:onEnd(function()
              st.dir = "r"
              spr.sprite:play("Idle_R")
              local tr = dwarf:getComponent("Transform")
              tr.pos.x = tr.pos.x + 48
            end)
            b:makeComponent("TileMovement", pos.pos.x+1, pos.pos.y, 1.082, 0.726)
            
            st.dir = "pr"
          end

        elseif self:isElevator(pos.pos.x + 1, pos.pos.y) then
          dwarf:makeComponent("TileMovement", pos.pos.x + 1, pos.pos.y, 1.3).ease = "linear"
        elseif not self:isFloor(pos.pos.x + 1, pos.pos.y) then
          if not self:isFloor(pos.pos.x + 1, pos.pos.y + 1) then
            if self:isFloor(pos.pos.x + 1, pos.pos.y + 2) then
              st.dir = "dr"
              pos.pos.y = pos.pos.y + 1
              pos.pos.x = pos.pos.x + 1
              spr.sprite:play("Drop_R")
              play_drop()
              spr.sprite:onEnd(function()
                st.dir = "r"
                spr.sprite:play("Idle_R")
                local tr = dwarf:getComponent("Transform")
                tr.pos.x = tr.pos.x +48
                tr.pos.y = tr.pos.y +64 
              end)

            elseif self:isFloor(pos.pos.x + 1, pos.pos.y + 3) then
              st.dir = "ddr"
              pos.pos.y = pos.pos.y + 2
              pos.pos.x = pos.pos.x + 1
              spr.sprite:play("Drop_Deep_R")
              self:deep_dust()
              G.sounds.deep:stop()
              G.sounds.deep:play()
              spr.sprite:onEnd(function()
                st.dir = "l"
                spr.sprite:play("Idle_L")
                local tr = dwarf:getComponent("Transform")
                tr.pos.x = tr.pos.x +48
                tr.pos.y = tr.pos.y +128 
              end)
            end
          else
            dwarf:makeComponent("TileMovement", pos.pos.x + 1, pos.pos.y, 1.3).ease = "linear"
          end
        end
      else
        st.dir = "tr"
        spr.sprite:play("Turn_L_R")
        spr.sprite:onEnd(function()
          st.dir = "r"
          spr.sprite:play("Idle_R")
        end)
      end

    elseif event == "left" then
      if st.dir == "l" then
        local b = self:isBlock(pos.pos.x - 1, pos.pos.y)
        if b then
          if not self:isFloor(pos.pos.x - 2, pos.pos.y) and not self:isBlock(pos.pos.x - 1, pos.pos.y-1) then
            -- PUSH LEFT
            spr.sprite:play("Push_L")
            G.sounds.push:stop()
            G.sounds.push:play()
            pos.pos.x = pos.pos.x - 1
            spr.sprite:onEnd(function()
              st.dir = "l"
              spr.sprite:play("Idle_L")
              local tr = dwarf:getComponent("Transform")
              tr.pos.x = tr.pos.x -16 
            end)
            b:makeComponent("TileMovement", pos.pos.x-1, pos.pos.y, 1.082, 0.726)
            st.dir = "pl"
          end

        elseif self:isElevator(pos.pos.x - 1, pos.pos.y) then
          dwarf:makeComponent("TileMovement", pos.pos.x - 1, pos.pos.y, 1.3).ease = "linear"

        elseif not self:isFloor(pos.pos.x - 1, pos.pos.y) then
          if not self:isFloor(pos.pos.x - 1, pos.pos.y + 1) then
            if self:isFloor(pos.pos.x - 1, pos.pos.y + 2) then
              st.dir = "dl"
              pos.pos.y = pos.pos.y + 1
              pos.pos.x = pos.pos.x - 1
              spr.sprite:play("Drop_L")
              play_drop()
              spr.sprite:onEnd(function()
                st.dir = "l"
                spr.sprite:play("Idle_L")
                local tr = dwarf:getComponent("Transform")
                tr.pos.x = tr.pos.x -16
                tr.pos.y = tr.pos.y +64 
              end)

            elseif self:isFloor(pos.pos.x - 1, pos.pos.y + 3) then
              st.dir = "ddl"
              pos.pos.y = pos.pos.y + 2
              pos.pos.x = pos.pos.x - 1
              G.sounds.deep:stop()
              G.sounds.deep:play()
              spr.sprite:play("Drop_Deep_L")
              self:deep_dust()
              spr.sprite:onEnd(function()
                st.dir = "r"
                spr.sprite:play("Idle_R")
                local tr = dwarf:getComponent("Transform")
                tr.pos.x = tr.pos.x -16
                tr.pos.y = tr.pos.y +128 
              end)

            
            end
          else
            dwarf:makeComponent("TileMovement", pos.pos.x - 1, pos.pos.y, 1.3).ease = "linear"
          end
        end
      else
        st.dir = "tl"
        spr.sprite:play("Turn_R_L")
        spr.sprite:onEnd(function()
          st.dir = "l"
          spr.sprite:play("Idle_L")
        end)
      end
    end
  end
end
